@@ -3,7 +3,7 @@ A custom graphic renderer for the '.plain' files produced by dot.
 
 """
 
-from __future__ import absolute_import
+from __future__ import absolute_import, print_function
 import re, os, math
 import pygame
 from pygame.locals import *
@@ -671,6 +671,16 @@ COLOR = {
     'yellow4': (139, 139, 0),
     'yellowgreen': (154, 205, 50),
     }
+
+GONS = {
+    b'triangle': 3,
+    b'diamond': 4,
+    b'pentagon': 5,
+    b'hexagon': 6,
+    b'septagon': 7,
+    b'octagon': 8,
+}
+
 re_nonword=re.compile(r'([^0-9a-zA-Z_.]+)')
 re_linewidth=re.compile(forcestr(r'setlinewidth\((\d+(\.\d*)?|\.\d+)\)'))
 
@@ -731,7 +741,7 @@ class GraphLayout:
         self.edges.append(Edge(self.nodes, *args))
 
     def get_display(self):
-        from graphdisplay import GraphDisplay
+        from dotviewer.graphdisplay import GraphDisplay
         return GraphDisplay(self)      
 
     def display(self):
@@ -914,6 +924,11 @@ def segmentdistance(p0, p1, p):
     else:
         return abs(vy*(x-x0) - vx*(y-y0))
 
+def ellipse(t, center, a, b):
+    """ compute points on an elliplse, with t running from 0 to 2*math.pi, a
+    and b being the radii and center as the origin of the ellipse. """
+    xcenter, ycenter = center
+    return int(xcenter - a / 2.0 * math.sin(t)), int(ycenter - b / 2.0 * math.cos(t))
 
 class GraphRenderer:
     MARGIN = 0.6
@@ -1049,6 +1064,8 @@ class GraphRenderer:
             if x-nw2 < w and x+nw2 > 0 and y-nh2 < h and y+nh2 > 0:
                 self.visiblenodes.append(node)
         for edge in self.graphlayout.edges:
+            if edge.style == b"invis":
+                continue
             x1, y1, x2, y2 = edge.limits()
             x1, y1 = self.map(x1, y1)
             if x1 < w and y1 < h:
@@ -1065,6 +1082,9 @@ class GraphRenderer:
                 self.bboxh - (py + (self.ofsy - self.margin)) / self.scale)
 
     def draw_node_commands(self, node):
+        if node.shape in (b"record", b'Mrecord'):
+            return self.draw_record_commands(node)
+
         xcenter, ycenter = self.map(node.x, node.y)
         boxwidth = int(node.w * self.scale)
         boxheight = int(node.h * self.scale)
@@ -1083,7 +1103,6 @@ class GraphRenderer:
         hmax = 0
         commands = []
         bkgndcommands = []
-
         if self.font is None:
             if lines:
                 raw_line = lines[0].replace('\\l','').replace('\r','')
@@ -1151,15 +1170,30 @@ class GraphRenderer:
         ytop = ycenter - hmax//2
         x = xcenter-boxwidth//2
         y = ycenter-boxheight//2
+        center = (xcenter, ycenter)
 
-        if node.shape == b'box':
+        if node.shape in (b'box', b'rect', b'rectangle'):
             rect = (x-1, y-1, boxwidth+2, boxheight+2)
+            if rect[0] < 0:
+                rect = (0, y-1, boxwidth+2 + x-1, boxheight+2)
+            if rect[1] < 0:
+                rect = (rect[0], 0, rect[2], boxheight+2 + y-1)
             def cmd():
                 self.screen.fill(bgcolor, rect)
             bkgndcommands.append(cmd)
             def cmd():
                 pygame.draw.rect(self.screen, fgcolor, rect, 1)
             commands.append(cmd)
+        elif node.shape == b'square':
+            width = max(boxwidth+2, boxheight+2)
+            rect = (x-1, y-1, width, width)
+            def cmd():
+                self.screen.fill(bgcolor, rect)
+            bkgndcommands.append(cmd)
+            def cmd():
+                pygame.draw.rect(self.screen, fgcolor, rect, 1)
+            commands.append(cmd)
+
         elif node.shape == b'ellipse':
             rect = (x-1, y-1, boxwidth+2, boxheight+2)
             def cmd():
@@ -1168,6 +1202,24 @@ class GraphRenderer:
             def cmd():
                 pygame.draw.ellipse(self.screen, fgcolor, rect, 1)
             commands.append(cmd)
+        elif node.shape == b'circle':
+            radius = max(boxwidth+2, boxheight+2) // 2
+            def cmd():
+                pygame.draw.circle(self.screen, bgcolor, center, radius, 0)
+            bkgndcommands.append(cmd)
+            def cmd():
+                pygame.draw.circle(self.screen, fgcolor, center, radius, 1)
+            commands.append(cmd)
+        elif node.shape == b'doublecircle':
+            radius = max(boxwidth+2, boxheight+2) // 2
+            bigradius = int(radius * 1.05)
+            def cmd():
+                pygame.draw.circle(self.screen, bgcolor, center, bigradius, 0)
+            bkgndcommands.append(cmd)
+            def cmd():
+                pygame.draw.circle(self.screen, fgcolor, center, radius, 1)
+                pygame.draw.circle(self.screen, fgcolor, center, bigradius, 1)
+            commands.append(cmd)
         elif node.shape == b'octagon':
             step = 1-math.sqrt(2)/2
             points = [(int(x+boxwidth*fx), int(y+boxheight*fy))
@@ -1181,6 +1233,43 @@ class GraphRenderer:
             def cmd():
                 pygame.draw.polygon(self.screen, fgcolor, points, 1)
             commands.append(cmd)
+        elif node.shape == b'doubleoctagon' or node.shape == b'tripleoctagon':
+            radius = max(boxwidth+2, boxheight+2)
+            # not quite right: not all sides are parallel
+            width = int(radius * 1.05) - radius
+            count = 8
+            points = [ellipse((2 * math.pi) / float(count) * (i + 0.5),
+                        center, boxwidth, boxheight)
+                      for i in range(count)]
+            points2 = [ellipse((2 * math.pi) / float(count) * (i + 0.5),
+                        center, boxwidth + width, boxheight + width)
+                      for i in range(count)]
+            points3 = [ellipse((2 * math.pi) / float(count) * (i + 0.5),
+                        center, boxwidth + 2 * width, boxheight + 2 * width)
+                      for i in range(count)]
+            if node.shape != b'tripleoctagon':
+                points3 = points2
+            def cmd():
+                pygame.draw.polygon(self.screen, bgcolor, points3, 0)
+            bkgndcommands.append(cmd)
+            def cmd():
+                pygame.draw.polygon(self.screen, fgcolor, points, 1)
+                pygame.draw.polygon(self.screen, fgcolor, points2, 1)
+                pygame.draw.polygon(self.screen, fgcolor, points3, 1)
+            commands.append(cmd)
+        elif node.shape in GONS:
+            count = GONS[node.shape]
+            points = [ellipse((2 * math.pi) / float(count) * i, center, boxwidth, boxheight)
+                      for i in range(count)]
+            def cmd():
+                pygame.draw.polygon(self.screen, bgcolor, points, 0)
+            bkgndcommands.append(cmd)
+            def cmd():
+                pygame.draw.polygon(self.screen, fgcolor, points, 1)
+            commands.append(cmd)
+
+        elif node.shape in (b'none', b'plain', b'plaintext'):
+            pass
         return bkgndcommands, commands
 
     def draw_commands(self):
@@ -1221,6 +1310,44 @@ class GraphRenderer:
 
         return edgebodycmd + nodebkgndcmd + edgeheadcmd + nodecmd
 
+    def draw_record_commands(self, node):
+        xcenter, ycenter = self.map(node.x, node.y)
+        boxwidth = int(node.w * self.scale)
+        boxheight = int(node.h * self.scale)
+        fgcolor = getcolor(node.color, (0,0,0))
+        bgcolor = getcolor(node.fillcolor, (255,255,255))
+        if node.highlight:
+            fgcolor = highlight_color(fgcolor)
+            bgcolor = highlight_color(bgcolor)
+
+        text = node.label
+
+        img = TextSnippetRecordLR.make(self, text, (boxwidth, boxheight), fgcolor, bgcolor)
+        wmax, hmax = img.get_size()
+        xleft = xcenter - boxwidth//2
+        ytop = ycenter - boxheight//2
+        def cmd(img=img, y=0):
+            img.draw(xleft, ytop+y)
+        commands = [cmd]
+        bkgndcommands = []
+
+        x = xcenter-boxwidth//2
+        y = ycenter-boxheight//2
+
+        # draw box
+        rect = (x-1, y-1, boxwidth+2, boxheight+2)
+        if rect[0] < 0:
+            rect = (0, y-1, boxwidth+2 + x-1, boxheight+2)
+        if rect[1] < 0:
+            rect = (rect[0], 0, rect[2], boxheight+2 + y-1)
+        def cmd():
+            self.screen.fill(bgcolor, rect)
+        bkgndcommands.append(cmd)
+        def cmd():
+            pygame.draw.rect(self.screen, fgcolor, rect, 1)
+        commands.append(cmd)
+        return bkgndcommands, commands
+
     def render(self):
         self.computevisible()
 
@@ -1362,3 +1489,144 @@ class TextSnippet:
             self.renderer.textzones.append((x, y, w, h, word))
             x += w
 
+
+def record_to_nested_lists(s):
+    curr = []
+    stack = []
+    curr_string = []
+    index = 0
+    prev = ''
+    while index < len(s):
+        char = s[index]
+        if char == "|":
+            if prev != '}':
+                curr.append("".join(curr_string))
+                curr_string = []
+        elif char == "{":
+            stack.append(curr)
+            curr = []
+        elif char == "}":
+            curr.append("".join(curr_string))
+            curr_string = []
+            stack[-1].append(curr)
+            curr = stack.pop()
+        elif char == "<":
+            index = s.find(">", index) + 1
+            if index == 0:
+                assert 0
+            else:
+                prev = ">"
+                continue
+        else:
+            curr_string.append(char)
+        index += 1
+        prev = char
+    if curr_string:
+        curr.append("".join(curr_string))
+    assert not stack
+    return curr
+
+
+class TextSnippetRecordLR:
+    def __init__(self, renderer, snippets, fgcolor, bgcolor=None, font=None):
+        if font is None:
+            font = renderer.font
+        if font is None:
+            return
+        self.snippets = snippets
+        self.renderer = renderer
+        self.fgcolor = fgcolor
+
+    @staticmethod
+    def make(renderer, s, bbox, fgcolor, bgcolor=None, font=None):
+        fgcolor = ensure_readable(fgcolor, bgcolor)
+        l = record_to_nested_lists(s)
+        res = TextSnippetRecordLR.make_from_list(renderer, l, fgcolor, bgcolor, font)
+        res._compute_offsets(bbox)
+        return res
+
+    @classmethod
+    def make_from_list(cls, renderer, l, fgcolor, bgcolor, font):
+        snippets = []
+        for element in l:
+            if isinstance(element, list):
+                snippets.append(cls.othercls.make_from_list(renderer, element, fgcolor, bgcolor, font))
+            else:
+                snippets.append(TextSnippet(renderer, element, fgcolor, bgcolor, font))
+        return cls(renderer, snippets, fgcolor, bgcolor, font)
+
+    def get_size(self):
+        if self.snippets:
+            sizes = [img.get_size() for img in self.snippets]
+            return sum([w for w,h in sizes]), max([h for w,h in sizes])
+        else:
+            return 0, 0
+
+    def draw(self, x, y):
+        for i, img in enumerate(self.snippets):
+            w, h = img.get_size()
+            if not isinstance(img, TextSnippetRecordLR):
+                img.draw(x + self.offsets[i] + self.spacing, y + (self.bbox[1] - h) // 2)
+            else:
+                img.draw(x + self.offsets[i], y)
+            if i != 0:
+                start = x + self.offsets[i], y
+                end = x + self.offsets[i], y + self.bbox[1]
+                pygame.draw.line(self.renderer.screen, self.fgcolor, start, end)
+
+    def _compute_offsets(self, bbox):
+        self.bbox = bbox
+        width_content, height_content = self.get_size()
+        width, height = bbox
+        spacing = (width - width_content) / 2.0 / len(self.snippets)
+        offset = 0
+        offsets = []
+        for i, img in enumerate(self.snippets):
+            w, h = img.get_size()
+            offsets.append(int(offset))
+            offset += w + 2 * spacing
+            if not isinstance(img, TextSnippet):
+                img._compute_offsets((w + 2 * spacing, height))
+        self.offsets = offsets
+        self.spacing = spacing
+
+class TextSnippetRecordTD(TextSnippetRecordLR):
+    othercls = TextSnippetRecordLR
+
+    def get_size(self):
+        if self.snippets:
+            sizes = [img.get_size() for img in self.snippets]
+            return max([w for w,h in sizes]), sum([h for w,h in sizes])
+        else:
+            return 0, 0
+
+    def draw(self, x, y):
+        for i, img in enumerate(self.snippets):
+            w, h = img.get_size()
+            if not isinstance(img, TextSnippetRecordLR):
+                img.draw(x + (self.bbox[0] - w) // 2, y + self.offsets[i] + self.spacing)
+            else:
+                img.draw(x, y + self.offsets[i])
+            if i != 0:
+                start = x, y + self.offsets[i]
+                end = x + self.bbox[0], y  + self.offsets[i]
+                pygame.draw.line(self.renderer.screen, self.fgcolor, start, end)
+
+    def _compute_offsets(self, bbox):
+        self.bbox = bbox
+        width_content, height_content = self.get_size()
+        width, height = bbox
+        spacing = (height - height_content) / 2.0 / len(self.snippets)
+        offset = 0
+        offsets = []
+        for i, img in enumerate(self.snippets):
+            w, h = img.get_size()
+            offsets.append(int(offset))
+            offset += h + 2 * spacing
+            if not isinstance(img, TextSnippet):
+                img._compute_offsets((width, h + 2 * spacing))
+        self.offsets = offsets
+        self.spacing = spacing
+
+
+TextSnippetRecordLR.othercls = TextSnippetRecordTD
